#!/usr/bin/env python3
# Kebechet
# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Common and useful utilities for managers."""

import logging
import platform
import typing
import git
import os

import delegator
import kebechet

from kebechet.exception import PipenvError
from thoth.sourcemanagement.enums import ServiceType
from thoth.sourcemanagement.sourcemanagement import SourceManagement


_LOGGER = logging.getLogger(__name__)


def _init_service_url(service_type: ServiceType = None, service_url: str = None) -> str:
    """Needed for the cron job to initialize the service_url if not provided."""
    if service_type == ServiceType.GITHUB:
        service_url = service_url or "https://github.com"
    elif service_type == ServiceType.GITLAB:
        service_url = service_url or "https://gitlab.com"
    else:
        raise NotImplementedError

    return service_url


class ManagerBase:
    """A base class for manager instances holding common and useful utilities."""

    def __init__(
        self,
        slug,
        service_type: ServiceType = None,
        service_url: str = None,
        parsed_payload: dict = None,
        token: str = None,
    ):
        """Initialize manager instance for talking to services."""
        self.service_type = service_type or ServiceType.GITHUB
        # This needs to be called before instantiation of service url, if not provided.
        self.service_url = _init_service_url(service_type, service_url)
        # Allow token expansion from env vars.
        self.slug = slug
        # Parsed payload structure can be accessed in payload_parser.py
        self.parsed_payload = None
        if parsed_payload:
            self.parsed_payload = parsed_payload
        self.owner, self.repo_name = self.slug.split("/", maxsplit=1)
        self.installation = False
        if os.getenv("GITHUB_PRIVATE_KEY_PATH") and os.getenv("GITHUB_APP_ID"):
            self.installation = True  # Authenticate as github app.
        self.sm = SourceManagement(
            service_type=self.service_type,
            service_url=self.service_url,
            token=token,
            slug=slug,
            installation=self.installation,
        )
        self._repo = None

    @property
    def repo(self):
        """Get repository on which we work on."""
        return self._repo

    @repo.setter
    def repo(self, repo: git.Repo):
        """Set repository information and all derived information needed."""
        self._repo = repo
        self.slug = repo.remote().url.split(":", maxsplit=1)[1][: -len(".git")]

    @classmethod
    def get_environment_details(
        cls, as_dict=False
    ) -> typing.Union[str, typing.Dict[str, str]]:
        """Get details for environment in which Kebechet runs."""
        try:
            pipenv_version = cls.run_pipenv("pipenv --version")
        except PipenvError as exc:
            pipenv_version = f"Failed to obtain pipenv version:\n{exc.stderr}"

        return (
            f"""
Kebechet version: {kebechet.__version__}
Python version: {platform.python_version()}
Platform: {platform.platform()}
pipenv version: {pipenv_version}
"""
            if not as_dict
            else {
                "kebechet_version": kebechet.__version__,
                "python_version": platform.python_version(),
                "platform": platform.platform(),
                "pipenv_version": pipenv_version,
            }
        )

    @staticmethod
    def run_pipenv(cmd: str):
        """Run pipenv, raise :ref:kebechet.exception.PipenvError on any error holding all the information."""
        _LOGGER.debug(f"Running pipenv command {cmd!r}")
        result = delegator.run(cmd)
        if result.return_code != 0:
            _LOGGER.warning(result.err)
            raise PipenvError(result)

        return result.out

    @classmethod
    def get_dependency_graph(cls, graceful: bool = False):
        """Get dependency graph of the project."""
        try:
            cls.run_pipenv("pipenv install --dev --skip-lock")
            return cls.run_pipenv("pipenv graph")
        except PipenvError as exc:
            if not graceful:
                raise
            return f"Unable to obtain dependency graph:\n\n{exc.stderr}"

    def run(self, labels: list) -> typing.Optional[dict]:
        """Run the given manager implementation."""
        raise NotImplementedError



#!/usr/bin/env python3
# Kebechet
# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Exceptions and errors that can be found in Kebechet."""

from delegator import Command


class KebechetException(Exception):
    """A base class for Kebechet exception hierarchy."""


class ConfigurationError(KebechetException):
    """Raised on error in configuration file."""


class PipenvError(KebechetException):
    """Raised on missing/invalid Pipenv or Pipenv.lock file."""

    def __init__(self, command: Command, *args, **kwargs):
        """Asssign values for exception so that they can be used in issue reports automatically."""
        self.command = command.cmd
        self.stdout = command.out
        self.stderr = command.err
        self.raw_command = command
        super().__init__(*args)


class InternalError(KebechetException):
    """Raised for internal errors, should not occur for end-user."""


class PullRequestError(KebechetException):
    """Raised in case of failed pull request."""


class DependencyManagementError(KebechetException):
    """An exception raised if there is an error in dependency management in the repo.

    This errors are usually wrong or missing Pipfile, Pipfile.lock, requirments.in or requirments.txt.
    """


class WebhookPayloadError(KebechetException):
    """An exception raised if there is an error if the webhook passed cannot be parsed as a json.

    Generally the payload is empty or not of proper json format.
    """


class CannotFetchPRError(KebechetException):
    """An exception raised if there OGR cannot fetch the PR's for a repo.

    Generally the token is invalid or doesn't have access to the repo.
    """


class CannotFetchBranchesError(KebechetException):
    """An exception raised if there OGR cannot fetch the repo branches.

    Generally the token is invalid or doesn't have access to the repo.
    """


class CreatePRError(KebechetException):
    """An exception raised if there OGR cannot create an PR.

    Generally the token is invalid or doesn't have access to the repo.
    """



#!/usr/bin/env python3
# Kebechet
# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""The Command Line Interface."""


import logging
import os

import click
import json
from thoth.common import init_logging
from kebechet.exception import WebhookPayloadError

from kebechet import __version__ as kebechet_version
from kebechet.config import config

init_logging(logging_env_var_start="KEBECHET_LOG_")

_LOGGER = logging.getLogger("kebechet")


def _print_version(ctx, _, value):
    """Print Kebechet version and exit."""
    if not value or ctx.resilient_parsing:
        return
    click.echo(kebechet_version)
    ctx.exit()


@click.group()
@click.pass_context
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    envvar="KEBECHET_VERBOSE",
    help="Be verbose about what's going on.",
)
@click.option(
    "--version",
    is_flag=True,
    is_eager=True,
    callback=_print_version,
    expose_value=False,
    help="Print version and exit.",
)
def cli(ctx=None, verbose=0):
    """CLI."""
    if ctx:
        ctx.auto_envvar_prefix = "KEBECHET"

    _LOGGER.info("Kebechet version: %r", kebechet_version)

    if verbose:
        _LOGGER.setLevel(logging.DEBUG)
        _LOGGER.debug("Debug mode turned on")


@cli.command("run")
@click.argument("configuration", metavar="config", envvar="KEBECHET_CONFIGURATION_PATH")
def cli_run(configuration):
    """Run Kebechet using provided YAML configuration file."""
    config.run(configuration)


@cli.command("run-results")
@click.option("-o", "--origin", envvar="KEBECHET_REPO_URL")
@click.option("-s", "--service", envvar="KEBECHET_SERVICE_NAME")
@click.option("-i", "--analysis_id", metavar="id", envvar="KEBECHET_ANALYSIS_ID")
def cli_run_results(origin, service, analysis_id):
    """Run Kebechet after results are received (meant to be triggered automatically)."""
    config.run_analysis(analysis_id=analysis_id, origin=origin, service=service)


@cli.command("run-url")
@click.option("-u", "--url", envvar="KEBECHET_REPO_URL")
@click.option("-s", "--service", envvar="KEBECHET_SERVICE_NAME")
def cli_run_url(url, service):
    """Run Kebechet by providing url to a git repository service."""
    config.run_url(url=url, service=service, parsed_payload=None, tls_verify=True)


@cli.command("run-webhook")
@click.argument("web_payload", nargs=1, envvar="KEBECHET_PAYLOAD")
def cli_run_webhook(web_payload):
    """Run Kebechet by providing a webhook payload."""
    payload = None
    if os.path.isfile(web_payload):
        with open(web_payload) as f:
            payload = json.load(f)
    else:
        # If the json is passed a string.
        payload = json.loads(web_payload)
    if not payload:
        raise WebhookPayloadError("Webhook payload is empty or cannot be parsed.")
    config.run_webhook(payload=payload)


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
# Kebechet
# Copyright(C) 2018, 2019, 2020 Kevin Postlethwait
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Consume Thoth Output for Kebechet auto-dependency management."""

import hashlib
import os
import logging
import json
import typing
from thamos import lib

import git  # noqa F401

from kebechet.exception import DependencyManagementError  # noqa F401
from kebechet.exception import InternalError  # noqa F401
from kebechet.exception import PipenvError  # noqa F401
from kebechet.managers.manager import ManagerBase
from thoth.sourcemanagement.sourcemanagement import Issue  # noqa F401
from thoth.sourcemanagement.sourcemanagement import PullRequest  # noqa F401
from kebechet.utils import cloned_repo
from thoth.common import ThothAdviserIntegrationEnum

_BRANCH_NAME = "kebechet_thoth"
_LOGGER = logging.getLogger(__name__)
# Github and Gitlab events on which the manager acts upon.
_EVENTS_SUPPORTED = ["push", "issues", "issue", "merge_request"]


class ThothAdviseManager(ManagerBase):
    """Manage updates of dependencies using Thoth."""

    def __init__(self, *args, **kwargs):
        """Initialize ThothAdvise manager."""
        # We do API calls once for merge requests and we cache them for later use.
        self._cached_merge_requests = None
        super().__init__(*args, **kwargs)

    @property
    def sha(self):
        """Get SHA of the current head commit."""
        return self.repo.head.commit.hexsha

    def _construct_branch_name(self) -> str:
        """Construct branch name for the updated dependency."""
        return f"{_BRANCH_NAME}-{self.sha[:10]}"

    def _git_push(
        self, commit_msg: str, branch_name: str, files: list, force_push: bool = False
    ) -> None:
        """Perform git push after adding files and giving a commit message."""
        self.repo.index.add(files)
        self.repo.index.commit(commit_msg)
        self.repo.remote().push(branch_name, force=force_push)

    def _open_merge_request(
        self, branch_name: str, labels: list, files: list
    ) -> typing.Optional[int]:
        """Open a pull/merge request for dependency update."""
        commit_msg = "Auto generated update"
        body = "Pipfile.lock updated by kebechet-thoth manager"

        # Delete branch if it didn't change Pipfile.lock
        diff = self.repo.git.diff("master", files)
        if diff == "":
            _LOGGER.info("No changes necessary, exiting...")
            return None

        # push force always to keep branch up2date with the recent master and avoid merge conflicts.
        _LOGGER.info("Pushing changes")
        self._git_push(":pushpin: " + commit_msg, branch_name, files, force_push=True)

        # Check if the merge request already exists
        for mr in self._cached_merge_requests:
            if mr.source_branch == branch_name:
                _LOGGER.info("Merge request already exists, updating...")
                return None

        _LOGGER.info("Opening merge request")
        merge_request = self.sm.open_merge_request(
            commit_msg, branch_name, body, labels
        )
        return merge_request

    @staticmethod
    def _write_advise(adv_results: list):
        lock_info = adv_results[0]["report"][0][1]["requirements_locked"]
        with open("Pipfile.lock", "w+") as f:
            _LOGGER.info("Writing to Pipfile.lock")
            _LOGGER.debug(f"{json.dumps(lock_info)}")
            f.write(json.dumps(lock_info))

    def _issue_advise_error(self, adv_results: list, labels: list):
        """Create an issue if advise fails."""
        _LOGGER.debug(json.dumps(adv_results))
        textblock = ""
        errors = adv_results[0]["report"][0][0]
        for error in errors:
            justification = error["justification"]
            type_ = error["type"]
            _LOGGER.info(f"Error type: {type_}")
            textblock = (
                textblock
                + f"## Error type: {type_}\n"
                + f"**Justification**: {justification}\n"
            )

        checksum = hashlib.md5(textblock.encode("utf-8")).hexdigest()[:10]
        _LOGGER.info("Creating issue")
        self.sm.open_issue_if_not_exist(
            f"{checksum} - Automated kebechet thoth-advise Issue",
            lambda: textblock,
            labels=labels,
        )

    def run(self, labels: list, analysis_id=None):
        """Run Thoth Advising Bot."""
        if self.parsed_payload:
            if self.parsed_payload.get("event") not in _EVENTS_SUPPORTED:
                _LOGGER.info(
                    "ThothAdviseManager doesn't act on %r events.",
                    self.parsed_payload.get("event"),
                )
                return

        if analysis_id is None:
            with cloned_repo(self.service_url, self.slug, depth=1) as repo:
                self.repo = repo
                if not os.path.isfile("Pipfile"):
                    _LOGGER.warning("Pipfile not found in repo... Creating issue")
                    self.sm.open_issue_if_not_exist(
                        "Missing Pipfile",
                        lambda: "Check your repository to make sure Pipfile exists",
                        labels=labels,
                    )
                    return False

                lib.advise_here(
                    nowait=True,
                    origin=(f"{self.service_url}/{self.slug}"),
                    source_type=ThothAdviserIntegrationEnum.KEBECHET,
                )
            return True
        else:
            with cloned_repo(self.service_url, self.slug, depth=1) as repo:
                self.repo = repo
                _LOGGER.info("Using analysis results from %s", analysis_id)
                res = lib.get_analysis_results(analysis_id)
                branch_name = self._construct_branch_name()
                branch = self.repo.git.checkout("-B", branch_name)  # noqa F841
                self._cached_merge_requests = self.sm.repository.get_pr_list()

                if res is None:
                    _LOGGER.error(
                        "Advise failed on server side, contact the maintainer"
                    )
                    return False
                _LOGGER.debug(json.dumps(res))

                if res[1] is False:
                    _LOGGER.info("Advise succeeded")
                    self._write_advise(res)
                    self._open_merge_request(branch_name, labels, ["Pipfile.lock"])
                    return True
                else:
                    _LOGGER.warning(
                        "Found error while running adviser... Creating issue"
                    )
                    self._issue_advise_error(res, labels)
                    return False



#!/usr/bin/env python3
# Kebechet
# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Body of issues automatically opened for information."""


INFO_REPORT = """
Information about dependency management for the current master branch with SHA {sha}.

<details>
  <summary>Dependency graph</summary>

```
{dependency_graph}
```

</details>

<details>
  <summary>Environment details</summary>

```
{environment_details}
```

</details>

"""



#!/usr/bin/env python3
# Kebechet
# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Just some utility methods."""


import os
import logging
from contextlib import contextmanager
from tempfile import TemporaryDirectory
from urllib.parse import urljoin

import git

from thoth.sourcemanagement.enums import ServiceType

_LOGGER = logging.getLogger(__name__)


@contextmanager
def cwd(path: str):
    """Change working directory in a push-pop manner with context manager."""
    previous_dir = os.getcwd()
    try:
        os.chdir(path)
        yield previous_dir
    finally:
        os.chdir(previous_dir)


@contextmanager
def cloned_repo(service_url: str, slug: str, **clone_kwargs):
    """Clone the given Git repository and cd into it."""
    if service_url.startswith("https://"):
        service_url = service_url[len("https://") :]
    elif service_url.startswith("http://"):
        service_url = service_url[len("http://") :]
    else:
        # This is mostly internal error - we require service URL to have protocol explicitly set
        raise NotImplementedError

    repo_url = f"git@{service_url}:{slug}.git"
    with TemporaryDirectory() as repo_path, cwd(repo_path):
        _LOGGER.info(f"Cloning repository {repo_url} to {repo_path}")
        repo = git.Repo.clone_from(repo_url, repo_path, branch="master", **clone_kwargs)
        repo.config_writer().set_value(
            "user", "name", os.getenv("KEBECHET_GIT_NAME", "Kebechet")
        ).release()
        repo.config_writer().set_value(
            "user",
            "email",
            os.getenv("KEBECHET_GIT_EMAIL", "noreply+kebechet@redhat.com"),
        ).release()
        yield repo


def construct_raw_file_url(
    service_url: str,
    slug: str,
    file_name: str,
    service_type: ServiceType,
    branch: str = None,
) -> str:
    """Get URL to a raw file - useful for downloads of content."""
    branch = branch or "master"
    if service_type == ServiceType.GITHUB:
        # TODO self hosted GitHub?
        url = f"https://raw.githubusercontent.com/{slug}/{branch}/{file_name}"
    elif service_type == ServiceType.GITLAB:
        url = urljoin(service_url, f"{slug}/raw/{branch}/{file_name}")
    else:
        raise NotImplementedError

    return url



#!/usr/bin/env python3
# Kebechet
# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Configuration Class."""

import logging
import os
import yaml
import urllib3
import requests
import typing

from .exception import ConfigurationError
from thoth.sourcemanagement.enums import ServiceType
from .services import Service
from .payload_parser import PayloadParser

_LOGGER = logging.getLogger(__name__)


class _Config:
    """Library-wide configuration."""

    def __init__(self):
        self._repositories = None

    def from_file(self, config_path: str):
        if config_path.startswith(("https://", "https://")):
            response = requests.get(config_path)
            response.raise_for_status()
            content = response.text
        else:
            with open(config_path) as config_file:
                content = config_file.read()

        try:
            self._repositories = yaml.safe_load(content).get("repositories") or []
        except Exception as exc:
            raise ConfigurationError(
                "Failed to parse configuration file: {str(exc)}"
            ) from exc

    @staticmethod
    def _managers_from_file(config_path: str):
        with open(config_path) as config_file:
            content = config_file.read()
        try:
            return yaml.safe_load(content).get("managers") or []
        except Exception as exc:
            raise ConfigurationError(
                "Failed to parse configuration file: {str(exc)}"
            ) from exc

    @staticmethod
    def _tls_verify_from_file(config_path: str):
        with open(config_path) as config_file:
            content = config_file.read()

        try:
            return yaml.safe_load(content).get("tls_verify") or False
        except Exception as exc:
            raise ConfigurationError(
                "Failed to parse configuration file: {str(exc)}"
            ) from exc

    @staticmethod
    def download_conf_from_url(url: str, service: str):
        _service_ = Service(service=service, url=url)
        tempfile = _service_.download_kebechet_config()
        return tempfile

    @classmethod
    def run_webhook(cls, payload: dict):
        _payload_ = PayloadParser(payload)
        parsed_payload = _payload_.parsed_data()
        if (
            parsed_payload
            and parsed_payload["url"] is not None
            and parsed_payload["service_type"] is not None
        ):
            cls.run_url(
                parsed_payload["url"],
                parsed_payload["service_type"],
                parsed_payload,
                True,
            )

    @classmethod
    def run_url(cls, url: str, service: str, parsed_payload: dict, tls_verify: bool):
        temp_file = cls.download_conf_from_url(url, service)
        _LOGGER.debug("Filename = %s", temp_file.name)

        global config
        from kebechet.managers import REGISTERED_MANAGERS

        tempfile = config.download_conf_from_url(url, service)
        managers = cls._managers_from_file(tempfile.name)

        # If called from run_webhook tls verify is always true.
        if tls_verify is None:
            tls_verify = cls._tls_verify_from_file(tempfile.name)

        scheme, _, host, _, slug, _, _ = urllib3.util.parse_url(url)
        slug = slug[1:]
        service_url = f"{scheme}://{host}"

        cls._tls_verification(service_url, slug, verify=tls_verify)

        if service_url and not service_url.startswith(("https://", "http://")):
            # We need to have this explicitly set for IGitt and also for security reasons.
            _LOGGER.error(
                "You have to specify protocol ('https://' or 'http://') in service URL "
                "configuration entry - invalid configuration %r",
                service_url,
            )

        if service_url and service_url.endswith("/"):
            service_url = service_url[:-1]

        _service_ = Service(service, url)
        token = _service_.token
        _LOGGER.debug("Using token %r%r", token[:3], "*" * len(token[3:]))

        for manager in managers:
            # We do pops on dict, which changes it. Let's create a soft duplicate so if a user uses
            # YAML references, we do not break.
            manager = dict(manager)
            try:
                manager_name = manager.pop("name")
            except Exception:
                _LOGGER.exception(
                    "No manager name provided in configuration entry for %r, ignoring entry",
                    slug,
                )
                continue
            kebechet_manager = REGISTERED_MANAGERS.get(manager_name)
            if not kebechet_manager:
                _LOGGER.error(
                    "Unable to find requested manager %r, skipping", manager_name
                )
                continue
            _LOGGER.info(f"Running manager %r for %r", manager_name, slug)
            manager_configuration = manager.get("configuration") or {}
            if manager:
                _LOGGER.warning(
                    "Ignoring option %r in manager entry for %r", manager, slug
                )
            try:
                instance = kebechet_manager(
                    slug, _service_.service, service_url, parsed_payload, token
                )
                instance.run(**manager_configuration)
            except Exception as exc:  # noqa F841
                _LOGGER.exception(
                    "An error occurred during run of manager %r %r for %r, skipping",
                    manager,
                    kebechet_manager,
                    slug,
                )

        temp_file.close()

    def iter_entries(self) -> typing.Iterable[typing.Tuple]:
        """Iterate over repositories listed."""
        for entry in self._repositories or []:
            try:
                items = dict(entry)
                value = (
                    items.get("managers") or [],
                    items.get("slug") or None,
                    items.get("service_type") or None,
                    items.get("service_url") or None,
                    items.get("token") or None,
                    items.get("tls_verify") or True,
                )

                if items:
                    _LOGGER.warning(
                        "Unknown configuration entry in configuration of %r: %r",
                        value[1],
                        items,
                    )

                yield value
            except KeyError:
                _LOGGER.exception(
                    "An error in your configuration - ignoring the given configuration entry..."
                )

    @staticmethod
    def _tls_verification(service_url: str, slug: str, verify: bool) -> None:
        """Turn off or on TLS verification based on configuration."""
        # We manage our own warnings, of course better ones!
        urllib3.disable_warnings()
        if not verify:
            _LOGGER.warning(
                "Turning off TLS certificate verification for %r hosted at %r",
                slug,
                service_url,
            )

        # Please close your eyes when reading this - it's pretty ugly solution but is somehow applicable to
        # the IGitt's handling of these methods.
        original_post = requests.Session.post
        original_delete = requests.Session.delete
        original_put = requests.Session.put
        original_get = requests.Session.get
        original_head = requests.Session.head
        original_patch = requests.Session.patch

        def post(*args, **kwargs):
            kwargs.pop("verify", None)
            return original_post(*args, **kwargs, verify=verify)

        requests.Session.post = post  # type: ignore

        def delete(*args, **kwargs):
            kwargs.pop("verify", None)
            return original_delete(*args, **kwargs, verify=verify)

        requests.Session.delete = delete  # type: ignore

        def put(*args, **kwargs):
            kwargs.pop("verify", None)
            return original_put(*args, **kwargs, verify=verify)

        requests.Session.put = put  # type: ignore

        def get(*args, **kwargs):
            kwargs.pop("verify", None)
            return original_get(*args, **kwargs, verify=verify)

        requests.Session.get = get  # type: ignore

        def head(*args, **kwargs):
            kwargs.pop("verify", None)
            return original_head(*args, **kwargs, verify=verify)

        requests.Session.head = head  # type: ignore

        def patch(*args, **kwargs):
            kwargs.pop("verify", None)
            return original_patch(*args, **kwargs, verify=verify)

        requests.Session.patch = patch  # type: ignore

    @classmethod
    def run_analysis(cls, analysis_id: str, origin: str, service: str) -> None:
        """Run result managers (meant to be triggered automatically)."""
        global config
        from kebechet.managers import ThothAdviseManager, ThothProvenanceManager

        tempfile = config.download_conf_from_url(origin, service)
        managers = _Config._managers_from_file(tempfile.name)
        tls_verify = _Config._tls_verify_from_file(tempfile.name)

        scheme, _, host, _, slug, _, _ = urllib3.util.parse_url(origin)
        slug = slug[1:]
        service_url = f"{scheme}://{host}"

        cls._tls_verification(service_url, slug, verify=tls_verify)

        if service_url and not service_url.startswith(("https://", "http://")):
            # We need to have this explicitly set for IGitt and also for security reasons.
            _LOGGER.error(
                "You have to specify protocol ('https://' or 'http://') in service URL "
                "configuration entry - invalid configuration %r",
                service_url,
            )

        if service_url and service_url.endswith("/"):
            service_url = service_url[:-1]

        _service_ = Service(service, origin)
        token = _service_.token
        _LOGGER.debug("Using token %r%r", token[:3], "*" * len(token[3:]))

        kebechet_manager: typing.Union[
            typing.Type[ThothAdviseManager], typing.Type[ThothProvenanceManager]
        ]
        for manager in managers:
            manager_name = manager.pop("name")
            if analysis_id.startswith("adviser") and manager_name == "thoth-advise":
                kebechet_manager = ThothAdviseManager
                break
            elif (
                analysis_id.startswith("provenance")
                and manager_name == "thoth-provenance"
            ):
                kebechet_manager = ThothProvenanceManager
                break
            else:
                _LOGGER.debug(
                    "Manager %r does not correspond with id:%r, skipping",
                    manager_name,
                    analysis_id,
                )
        else:
            _LOGGER.error("No manager configuration found for id: %r", analysis_id)
            return

        manager_config = manager.get("configuration") or {}
        manager_config["analysis_id"] = analysis_id
        # TODO: Fail if users add config entries not relative to given manager (open an issue)
        instance = kebechet_manager(slug, _service_.service, service_url, token)
        instance.run(**manager_config)

    @classmethod
    def run(cls, configuration_file: str) -> None:
        """Run Kebechet using provided YAML configuration file."""
        global config
        from kebechet.managers import REGISTERED_MANAGERS

        config.from_file(configuration_file)

        for (
            managers,
            slug,
            service_type,
            service_url,
            token,
            tls_verify,
        ) in config.iter_entries():
            cls._tls_verification(service_url, slug, verify=tls_verify)

            if service_url and not service_url.startswith(("https://", "http://")):
                # We need to have this explicitly set for IGitt and also for security reasons.
                _LOGGER.error(
                    "You have to specify protocol ('https://' or 'http://') in service URL "
                    "configuration entry - invalid configuration {service_url!}"
                )
                continue

            if service_url and service_url.endswith("/"):
                service_url = service_url[:-1]

            if token:
                # Allow token expansion based on env variables.
                token = token.format(**os.environ)
                _LOGGER.debug("Using token '%r%r'", token[:3], "*" * len(token[3:]))

            for manager in managers:
                # We do pops on dict, which changes it. Let's create a soft duplicate so if a user uses
                # YAML references, we do not break.
                manager = dict(manager)
                try:
                    manager_name = manager.pop("name")
                except Exception:
                    _LOGGER.exception(
                        "No manager name provided in configuration entry for %r, ignoring entry",
                        slug,
                    )
                    continue

                kebechet_manager = REGISTERED_MANAGERS.get(manager_name)
                if not kebechet_manager:
                    _LOGGER.error(
                        "Unable to find requested manager %r, skipping", manager_name
                    )
                    continue

                _LOGGER.info(f"Running manager %r for %r", manager_name, slug)
                manager_configuration = manager.get("configuration") or {}
                if manager:
                    _LOGGER.warning(
                        "Ignoring option %r in manager entry for %r", manager, slug
                    )

                try:
                    instance = kebechet_manager(
                        # The service type is set by default to github.
                        slug,
                        ServiceType.by_name(service_type),
                        service_url,
                        None,
                        token,
                    )
                    instance.run(**manager_configuration)
                except Exception as exc:  # noqa F841
                    _LOGGER.exception(
                        "An error occurred during run of manager %r %r for %r, skipping",
                        manager,
                        kebechet_manager,
                        slug,
                    )
            _LOGGER.info("Finished management for %r", slug)


config = _Config()



"""Tests for version manager."""

import pytest

from unittest.mock import patch

from git import Repo  # noqa F401
from thoth.sourcemanagement.enums import ServiceType
from kebechet.managers import VersionManager


class FakeRepo(object):
    """Repo class for testing purposes."""

    def __init__(self, *args, **kwargs):
        """Initialize FakeRepo instance."""
        self.git = FakeGit


class FakeGit(object):
    """Git class for testing purposes."""

    @staticmethod
    def log(*args, **kwargs) -> str:
        """Return fake logs."""
        return "Fake Log entry"

    @staticmethod
    def rev_list(*args, maxparents=0, **kwargs) -> str:
        """Return Fake rev list."""
        return "\n".join(["rev1", "rev2", "rev3", "rev4", "rev5"][: (maxparents + 1)])

    @staticmethod
    def tag(*args, **kwargs) -> str:
        """Return Fake repository tags."""
        return "\n".join(["0.1.0", "v0.2.0", "v1.0.0"])


class TestVersionManager:
    """Test version manager."""

    manager = VersionManager(
        slug="fake-user/fake-repo",
        service_type=ServiceType.GITHUB,
        token="test-token-xxx",
    )

    @pytest.mark.parametrize(
        "old_version,new_version,tag",
        [
            ("0.1.0", "0.2.0", "0.1.0"),
            ("0.2.0", "0.3.0", "v0.2.0"),
            ("1.1.0", "1.2.0", "rev1"),
        ],
    )
    def test__compute_changelog(self, old_version, new_version, tag):
        """Test VersionManager._compute_changelog static method."""
        # check that tag and version are matched correctly
        fake_repo = FakeRepo()

        with patch.object(FakeGit, "log") as patch_log:
            _ = self.manager._compute_changelog(
                repo=fake_repo,
                old_version=old_version,
                new_version=new_version,
                version_file=False,
                changelog_smart=False,
                changelog_classifier=None,
                changelog_format=None,
            )
            patch_log.assert_called_with(f"{tag}..HEAD", no_merges=True, format="* %s")



#!/usr/bin/env python3
# Kebechet
# Copyright(C) 2018, 2019, 2020 Kevin Postlethwait
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Consume Thoth Output for Kebechet auto-dependency management."""

import hashlib
import os
import logging
import pprint

from thamos import lib
import git  # noqa F401

from kebechet.exception import DependencyManagementError  # noqa F401
from kebechet.exception import InternalError  # noqa F401
from kebechet.exception import PipenvError  # noqa F401
from kebechet.managers.manager import ManagerBase
from thoth.sourcemanagement.sourcemanagement import Issue  # noqa F401
from thoth.sourcemanagement.sourcemanagement import PullRequest  # noqa F401
from kebechet.utils import cloned_repo


_LOGGER = logging.getLogger(__name__)

_BRANCH_NAME = "kebechet_thoth"
# Github and Gitlab events on which the manager acts upon.
_EVENTS_SUPPORTED = ["push", "issues", "issue", "merge_request"]


class ThothProvenanceManager(ManagerBase):
    """Manage source issues of dependencies."""

    def __init__(self, *args, **kwargs):
        """Initialize ThothProvenance manager."""
        self._cached_merge_requests = None
        super().__init__(*args, **kwargs)

    def _issue_provenance_error(self, prov_results: list, labels: list):
        error_info = prov_results[0]
        text_block = ""

        for err in error_info:
            _LOGGER.info(f"{err['id']}: {err['package_name']} {err['package_version']}")
            text_block = (
                text_block
                + f"## {err['type']}: {err['id']} - {err['package_name']}:{err['package_version']}\n"
                + f"**Justification: {err['justification']}**\n"
                + f"#### source: \n {pprint.pformat(err['source'])}\n"
                + f"#### lock_info:\n {pprint.pformat(err['package_locked'])}"
            )

        checksum = hashlib.md5(text_block.encode("utf-8")).hexdigest()[:10]

        self.sm.open_issue_if_not_exist(
            f"{checksum} - {len(error_info)}: Automated kebechet thoth-provenance Issue",
            lambda: text_block,
            labels=labels,
        )

    def run(self, labels: list, analysis_id=None):
        """Run the provenance check bot."""
        if self.parsed_payload:
            if self.parsed_payload.get("event") not in _EVENTS_SUPPORTED:
                _LOGGER.info(
                    "ThothProvenanceManager doesn't act on %r events.",
                    self.parsed_payload.get("event"),
                )
                return

        if not analysis_id:
            with cloned_repo(self.service_url, self.slug, depth=1) as repo:
                self.repo = repo
                if not (os.path.isfile("Pipfile") and os.path.isfile("Pipfile.lock")):
                    _LOGGER.warning(
                        "Pipfile or Pipfile.lock is missing from repo, opening issue"
                    )
                    self.sm.open_issue_if_not_exist(
                        "Missing pipenv files",
                        lambda: "Check your repository to make sure Pipfile and Pipfile.lock exist.",
                        labels=labels,
                    )
                    return False
                _LOGGER.info((self.service_url + self.slug))
                lib.provenance_check_here(
                    nowait=True, origin=f"{self.service_url}/{self.slug}"
                )
            return True
        else:
            with cloned_repo(self.service_url, self.slug, depth=1) as repo:
                res = lib.get_analysis_results(analysis_id)
                if res is None:
                    _LOGGER.error(
                        "Provenance check failed on server side, contact the maintainer"
                    )
                    return False
                if res[1] is False:
                    _LOGGER.info("Provenance check found problems, creating issue...")
                    self._issue_provenance_error(res, labels)
                    return False
                else:
                    _LOGGER.info(
                        "Provenance check found no problems, carry on coding :)"
                    )
                    return True



#!/usr/bin/env python3
# Kebechet
# Copyright(C) 2018, 2019, 2020 Kevin Postlethwait
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Provides abstraction of service utility functions."""


import requests
import tempfile
import logging
import os

from thoth.sourcemanagement.enums import ServiceType

_LOGGER = logging.getLogger(__name__)


class Service:
    """Allows for abstraction of multiple different service with easy support of my by changing `services`."""

    @staticmethod
    def _keep_slug(slug: str):
        return slug

    @staticmethod
    def _encode_fslash(slug: str):
        return slug.replace("/", "%2F")

    # NOTE: config_url could be generated using construct_raw_file_url but I feel this makes the code easier to follow
    _SERVICES = {
        "github": {
            "token": "GITHUB_KEBECHET_TOKEN",
            "service_type": ServiceType.GITHUB,
            "config_url": "https://raw.githubusercontent.com/{slug}/{branch}/.thoth.yaml",
            "auth": {"header": "Authorization", "value": "token {token}"},
            "slug_method": _keep_slug.__func__,  # type: ignore
        },
        "gitlab": {
            "token": "GITLAB_KEBECHET_TOKEN",
            "service_type": ServiceType.GITLAB,
            "config_url": "https://gitlab.com/api/v4/projects/{slug}/repository/files/.thoth.yaml/raw?ref={branch}",
            "auth": {"header": "Private-Token", "value": "{token}"},
            "slug_method": _encode_fslash.__func__,  # type: ignore
        },
    }

    def __init__(self, service, url, branch="master"):
        """Initialize a Service object, calls constructor functions `slug_method` and `_generate_slug`."""
        if service not in self._SERVICES.keys():
            raise ValueError(f"{service} is not supported at this time")
        self.service = self._SERVICES[service]["service_type"]
        _LOGGER.info("%s service detected", service)
        self.token = os.environ[self._SERVICES[service]["token"]]
        self.branch = branch
        self.url = url
        self.service_info = self._SERVICES[service]
        slug = Service._generate_slug(url)
        self.slug = self._SERVICES[service]["slug_method"](slug)

    @staticmethod
    def _generate_slug(url: str):
        params = url.split("/")
        return "/".join(params[3:])

    def get_kebechet_download_url(self):
        """Get the download url of the .kebechet.yaml."""
        return self.service_info["config_url"].format(
            slug=self.slug, branch=self.branch
        )

    def get_auth_header(self):
        """Get the authorization header for this services api."""
        return {
            self.service_info["auth"]["header"]: self.service_info["auth"][
                "value"
            ].format(token=self.token)
        }

    def download_kebechet_config(self):
        """Return a temporary file containing this repos' .thoth.yaml contents."""
        down_url = self.get_kebechet_download_url()
        auth = self.get_auth_header()
        _LOGGER.info("Downloading from %s", down_url)
        resp = requests.get(down_url, headers=auth)
        file_ = tempfile.NamedTemporaryFile()
        file_.write(resp.content)
        file_.seek(0)
        _LOGGER.info(resp.content)
        return file_



#!/usr/bin/env python3
# Kebechet
# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Body of issues and pull requests automatically opened."""

# A base information for update that failed with all the relevant information.
PIPENV_REPORT = """
##### Command

```
  $ {command}
```

<details>
  <summary>Standard output</summary>

```
{stdout}
```

</details>

<details>
  <summary>Standard error</summary>

```
{stderr}
```

</details>

<details>
  <summary>Environment details</summary>

```
{environment_details}
```

</details>

<details>
  <summary>Dependency graph</summary>

```
{dependency_graph}
```

</details>

"""

# Issued when updating all dependencies fails.
ISSUE_PIPENV_UPDATE_ALL = (
    """
Automatic dependency update failed for the current master with SHA {sha}.

The automatic dependency management cannot continue. Please fix errors reported bellow.

"""
    + PIPENV_REPORT
    + """

##### Notes

For more information, see [Pipfile]({pip_url}) and [Pipfile.lock]({piplock_url}).

Once this issue is resolved, the issue will be automatically closed by bot.

"""
)

# A refresh comment when master branch changed when updating all dependencies (issue was already created).
ISSUE_COMMENT_UPDATE_ALL = (
    """
Automatic dependency update still failing for the current master with SHA {sha}.
"""
    + PIPENV_REPORT
)


# A close comment when update all works again.
ISSUE_CLOSE_COMMENT = """
Closing this issue as it is no longer relevant for the current master with SHA {sha}.
"""


# Issue created when the environment cannot be replicated.
ISSUE_REPLICATE_ENV = """
Unable to replicate environment provided in [Pipfile.lock]({piplock_url}).
Linked SHA - {sha}

Most likely the deployment build will fail.

##### Command

```
  $ {command}
```

<details>
  <summary>Standard output</summary>

```
{stdout}
```

</details>

<details>
  <summary>Standard error</summary>

```
{stderr}
```

</details>

<details>
  <summary>Environment details</summary>

```
{environment_details}
```

</details>

For more information, see [Pipfile]({pip_url}) and [Pipfile.lock]({piplock_url}).
"""

ISSUE_NO_DEPENDENCY_MANAGEMENT = """No dependency management found for this repository. If you want to keep your dependencies managed, \
please submit `Pipfile` or `requirements.in` or `requirements-dev.in` file.

To generate a `Pipfile`, use:
```
$ pipenv install --skip-lock --code ./
$ git add Pipfile
$ git commit -m 'Add Pipfile for dependency management'
```

Make sure your `Pipfile` or `requirements.in` or `requirements-dev.in` is placed in the root of your Git repository.
"""


ISSUE_INITIAL_LOCK = """Failed to perform initial lock of your dependencies based on your [{file}]({url}}).
       Linked SHA - {sha}

See attached report below to inspect this issue.

##### Command

```
  $ {command}
```

<details>
  <summary>Standard output</summary>

```
{stdout}
```

</details>

<details>
  <summary>Standard error</summary>

```
{stderr}
```

</details>

<details>
  <summary>Environment details</summary>

```
{environment_details}
```
</details>
"""


ISSUE_UNSUPPORTED_PACKAGE = """Kebechet cannot support maintaining this application as it contain's Git version of packages.

  The package causing the issue is - {package}
  Linked SHA - {sha}

  For more information, see [Pipfile]({pip_url}) and [Pipfile.lock]({piplock_url}).

  <details>
    <summary>Environment details</summary>

  ```
  {environment_details}
  ```
  </details>
  """



#!/usr/bin/env python3
# Kebechet
# Copyright(C) 2020 Sai Sankar Gochhayat
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Provides abstraction of payload parsing functions."""

from thoth.sourcemanagement.enums import ServiceType  # noqa F401
import logging
from .exception import WebhookPayloadError
from urllib.parse import urlparse
from typing import Optional, Any, Dict

_LOGGER = logging.getLogger(__name__)


class PayloadParser:
    """Allow Kebechet to parse webhook payload of different services."""

    _GITHUB = "api.github.com"
    _GITLAB = "gitlab.com"

    _IGNORED_GITHUB_EVENTS = ["installation", "integration_installation"]

    def __init__(self, payload: dict) -> None:
        """Initialize the parameters we require from the services."""
        self.raw_payload = payload
        self.service_type = None
        self.url = None
        self.event = None
        self.parsed_payload: Optional[dict] = None

        # For github webhooks
        if "event" in payload:
            if payload["event"] not in self._IGNORED_GITHUB_EVENTS:
                github_payload = payload["payload"]
                parsed_url = urlparse(github_payload["sender"]["url"])
                if self._GITHUB in parsed_url.netloc:
                    self.service_type = "github"
                    self.event = payload["event"]
                    self.github_parser(github_payload)
        # For gitlab webhooks
        elif "payload" in payload:
            gitlab_payload = payload["payload"]
            parsed_url = urlparse(gitlab_payload["project"]["web_url"])
            if self._GITLAB in parsed_url.netloc:
                self.service_type = "gitlab"
                self.gitlab_parser(gitlab_payload)
        else:
            raise WebhookPayloadError("Payload passed is not supported.")

    def github_parser(self, payload: dict) -> None:
        """Parse Github data."""
        self.url = payload["repository"]["html_url"]

    def gitlab_parser(self, payload: dict) -> None:
        """Parse Gitlab data."""
        self.url = payload["project"]["web_url"]
        self.event = payload["object_kind"]

    def parsed_data(self) -> Optional[Dict[str, Optional[Any]]]:
        """Return the parsed data if its of a supported service."""
        if not self.service_type:
            return None
        self.parsed_payload = {
            "service_type": self.service_type,
            "url": self.url,
            "event": self.event,
            "raw_payload": self.raw_payload,
        }
        return self.parsed_payload



"""Update dependencies stated in the Pipfile.lock for the given repo or repositories."""

from thoth.sourcemanagement.enums import ServiceType  # noqa F401
from .managers import InfoManager  # noqa F401
from .managers import UpdateManager  # noqa F401

__name__ = "kebechet"
__version__ = "1.0.10"
__author__ = "Fridolin Pokorny <fridolin.pokorny@gmail.com>"



"""This file contains packaging information for kebechet module."""
import os
from setuptools import setup, find_packages
from pathlib import Path


def get_install_requires():  # noqa: D103
    with open("requirements.txt", "r") as requirements_file:
        # TODO: respect hashes in requirements.txt file
        res = requirements_file.readlines()
        return [req.split(" ", maxsplit=1)[0] for req in res if req]


def get_version():  # noqa: D103
    with open(os.path.join("kebechet", "__init__.py")) as f:
        content = f.readlines()

    for line in content:
        if line.startswith("__version__ ="):
            # dirty, remove trailing and leading chars
            return line.split(" = ")[1][1:-2]
    raise ValueError("No version identifier found")


setup(
    name="kebechet",
    entry_points={"console_scripts": ["kebechet=kebechet.cli:cli"]},
    version=get_version(),
    description="Keep your dependencies in your projects fresh and up2date",
    long_description=Path("README.rst").read_text(),
    author="Fridolin Pokorny",
    author_email="fridolin@redhat.com",
    license="GPLv3+",
    packages=find_packages(),
    install_requires=get_install_requires(),
)



#!/usr/bin/env python3
# Kebechet
# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Report information about repository and Kebechet itself."""

import logging
import typing

from kebechet.managers.manager import ManagerBase
from kebechet.utils import cloned_repo

from .messages import INFO_REPORT

_INFO_ISSUE_NAME = "Kebechet info"

_LOGGER = logging.getLogger(__name__)
# Github and Gitlab events on which the manager acts upon.
_EVENTS_SUPPORTED = ["issues", "issue"]


class InfoManager(ManagerBase):
    """Manager for submitting information about running Kebechet instance."""

    def run(self) -> typing.Optional[dict]:  # type: ignore
        """Check for info issue and close it with a report."""
        if self.parsed_payload:
            if self.parsed_payload.get("event") not in _EVENTS_SUPPORTED:
                _LOGGER.info(
                    "Info manager doesn't act on %r events.",
                    self.parsed_payload.get("event"),
                )
                return None
        issue = self.sm.get_issue(_INFO_ISSUE_NAME)
        if not issue:
            _LOGGER.info("No issue to report to, exiting")
            return None

        _LOGGER.info(f"Found issue {_INFO_ISSUE_NAME}, generating report")
        with cloned_repo(self.service_url, self.slug, depth=1) as repo:
            # We could optimize this as the get_issue() does API calls as well. Keep it this simple now.
            self.sm.close_issue_if_exists(
                _INFO_ISSUE_NAME,
                INFO_REPORT.format(
                    sha=repo.head.commit.hexsha,
                    slug=self.slug,
                    environment_details=self.get_environment_details(),
                    dependency_graph=self.get_dependency_graph(graceful=True),
                ),
            )
        return None



"""Managers implemented in Kebechet."""

from .info import InfoManager
from .update import UpdateManager
from .version import VersionManager  # noqa F401
from .pipfile_requirements import PipfileRequirementsManager
from .thoth_advise import ThothAdviseManager
from .thoth_provenance import ThothProvenanceManager

REGISTERED_MANAGERS = {
    "update": UpdateManager,
    "info": InfoManager,
    "version": VersionManager,
    "pipfile-requirements": PipfileRequirementsManager,
    "thoth-advise": ThothAdviseManager,
    "thoth-provenance": ThothProvenanceManager,
}



#!/usr/bin/env python3
# Kebechet
# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Automatically issue a new PR with adjusted version for Python projects."""

import os
import re
import typing

import logging

from git import Repo
from ogr.abstract import Issue
import yaml
import semver
from datetime import datetime

from kebechet.utils import cloned_repo
from kebechet.managers.manager import ManagerBase
from thoth.glyph import generate_log
from thoth.glyph import MLModel
from thoth.glyph import Format
from thoth.glyph import ThothGlyphException

_LOGGER = logging.getLogger(__name__)
_VERSION_PULL_REQUEST_NAME = "Release of version {}"
_NO_VERSION_FOUND_ISSUE_NAME = (
    f"No version identifier found in sources to perform a release"
)
_MULTIPLE_VERSIONS_FOUND_ISSUE_NAME = (
    f"Multiple version identifiers found in sources to perform a new release."
)
_MULTIPLE_VERSIONS_FOUND_ISSUE_BODY = f"Please have only one version string, to facilitate automated releases.\
         Multiple version strings found in these files - "
_NO_MAINTAINERS_ERROR = "No release maintainers stated for this repository"
_BODY_TRUNCATED = "The changelog body was truncated, please check CHANGELOG.md for the complete changelog."
_DIRECT_VERSION_TITLE = " release"
_RELEASE_TITLES = {
    "new calendar release": lambda _: datetime.utcnow().strftime("%Y.%m.%d"),
    "new major release": lambda current_version: semver.VersionInfo.parse(
        current_version
    ).bump_major(),
    "new minor release": lambda current_version: semver.VersionInfo.parse(
        current_version
    ).bump_minor(),
    "new patch release": lambda current_version: semver.VersionInfo.parse(
        current_version
    ).bump_patch(),
    "new pre-release": lambda current_version: semver.VersionInfo.parse(
        current_version
    ).bump_prerelease(),
    "new build release": lambda current_version: semver.VersionInfo.parse(
        current_version
    ).bump_build(),
    "finalize version": lambda current_version: semver.VersionInfo.parse(
        current_version
    ).finalize_version(),
}

# Github and Gitlab events on which the manager acts upon.
_EVENTS_SUPPORTED = ["issues", "issue"]
# Maximum number of log messages in a single release. Set due to ultrahook limits.
_MAX_CHANELOG_SIZE = 300


class VersionError(Exception):
    """An exception raised on invalid version provided or found in the repo."""


class VersionManager(ManagerBase):
    """Automatic version management for Python projects."""

    def _adjust_version_file(
        self, file_path: str, issue: Issue
    ) -> typing.Optional[tuple]:
        """Adjust version in the given file, return signalizes whether the return value indicates change in file."""
        with open(file_path, "r") as input_file:
            content = input_file.read().splitlines()

        changed = False
        new_version = None
        old_version = None
        for idx, line in enumerate(content):
            if line.startswith("__version__ = "):
                parts = line.split(" = ", maxsplit=1)
                if len(parts) != 2:
                    _LOGGER.warning(
                        "Found '__version__' identifier but unable to parse old version, skipping: %r",
                        line,
                    )
                    continue

                old_version = parts[1][1:-1]  # Remove ' and " in string representation.
                _LOGGER.info("Old version found in sources: %s", old_version)

                new_version = self._get_new_version(issue.title.strip(), old_version)
                _LOGGER.info("Computed new version: %s", new_version)

                content[idx] = f'__version__ = "{new_version}"'
                changed = True

        if not changed:
            return None

        # Apply changes.
        with open(file_path, "w") as output_file:
            output_file.write("\n".join(content))
            # Add new line at the of file explicitly.
            output_file.write("\n")

        return new_version, old_version

    def _adjust_version_in_sources(
        self, repo: Repo, labels: typing.Optional[list], issue: Issue
    ) -> typing.Optional[tuple]:
        """Walk through the directory structure and try to adjust version identifier in sources."""
        adjusted = []
        for root, _, files in os.walk("./"):
            for file_name in files:
                if file_name in (
                    "setup.py",
                    "__init__.py",
                    "__about__.py",
                    "version.py",
                    "app.py",
                    "wsgi.py",
                ):
                    file_path = os.path.join(root, file_name)
                    adjusted_version = self._adjust_version_file(file_path, issue)
                    if adjusted_version:
                        repo.git.add(file_path)
                        adjusted.append(
                            (file_path, adjusted_version[0], adjusted_version[1])
                        )

        if len(adjusted) == 0:
            error_msg = _NO_VERSION_FOUND_ISSUE_NAME
            _LOGGER.warning(error_msg)
            self.sm.open_issue_if_not_exist(
                error_msg,
                lambda: "Automated version release cannot be performed.\nRelated: #"
                + str(issue.id),
                None,
                labels,
            )

        if len(adjusted) > 1:
            error_msg = _MULTIPLE_VERSIONS_FOUND_ISSUE_NAME
            file_locations = [loc for loc, _, _ in adjusted]
            _LOGGER.warning(error_msg)
            _issue_body = (
                _MULTIPLE_VERSIONS_FOUND_ISSUE_BODY
                + "\n`"
                + ", ".join(file_locations)
                + "`"
            )
            self.sm.open_issue_if_not_exist(
                error_msg,
                lambda: _issue_body + "\nRelated: #" + str(issue.id),
                None,
                labels,
            )

        # Return old and new version identifier.
        return adjusted[0][1], adjusted[0][2]

    def _get_maintainers(self, labels: list = None) -> list:
        """Get maintainers based on configuration.

        Maintainers can be either stated in the configuration or in the OWNERS file in the repo itself.
        """
        try:
            with open("OWNERS", "r") as owners_file:
                owners = yaml.safe_load(owners_file)
            maintainers = list(map(str, owners.get("maintainers") or []))
        except (FileNotFoundError, KeyError, ValueError, yaml.YAMLError):
            _LOGGER.exception("Failed to load maintainers file")
            self.sm.open_issue_if_not_exist(
                _NO_MAINTAINERS_ERROR,
                lambda: "This repository is not correctly setup for automated version releases. "
                "Please revisit bot configuration.",
                labels=labels,
            )
            return []

        self.sm.close_issue_if_exists(
            _NO_MAINTAINERS_ERROR, "No longer relevant for the current bot setup."
        )
        return maintainers

    @staticmethod
    def _get_new_version(
        issue_title: str, current_version: str
    ) -> typing.Optional[str]:
        """Get next version based on user request."""
        issue_title = issue_title.lower()

        handler = _RELEASE_TITLES.get(issue_title)
        if handler:
            try:
                return str(handler(current_version))
            except ValueError as exc:  # Semver raises ValueError when version cannot be parsed.
                raise VersionError(
                    f"Wrong version specifier found in sources: `{current_version}`"
                ) from exc

        if issue_title.endswith(_DIRECT_VERSION_TITLE):  # a specific release
            parts = issue_title.split(" ")
            if len(parts) == 2:
                return parts[0]

        return None

    @staticmethod
    def _is_release_request(issue_title):
        """Check for possible candidate for a version bump."""
        issue_title = issue_title.lower()
        return (
            _RELEASE_TITLES.get(issue_title) is not None
            or issue_title.endswith(_DIRECT_VERSION_TITLE)
            and len(issue_title.split(" ")) == 2
        )

    @staticmethod
    def _compute_changelog(
        repo: Repo,
        old_version: str,
        new_version: str,
        changelog_smart: bool,
        changelog_classifier: str,
        changelog_format: str,
        version_file: bool = False,
    ) -> typing.List[str]:
        """Compute changelog for the given repo.

        If version file is used, add changelog to the version file and add changes to git.
        """
        _LOGGER.info(
            "Computing changelog for new release from version %r to version %r",
            old_version,
            new_version,
        )

        tags = repo.git.tag().splitlines()

        is_tagged_version = False
        for tag in tags:
            if old_version == tag or re.match(f"v?{old_version}", tag):
                old_version = tag
                is_tagged_version = True
                break

        if not is_tagged_version:
            _LOGGER.info(
                "Old version was not found in the git tag history, assuming initial release"
            )
            # Use the initial commit if this the previous tag was not found - this
            # can be in case of the very first release.
            old_version = repo.git.rev_list("HEAD", max_parents=0)

        _LOGGER.info("Smart Log : %s", str(changelog_smart))

        if changelog_smart:
            _LOGGER.info("Classifier : %s", changelog_classifier)
            _LOGGER.info("Format : %s", changelog_format)
            changelog = repo.git.log(
                f"{old_version}..HEAD", no_merges=True, format="%s"
            ).splitlines()
            changelog = generate_log(
                changelog,
                Format.by_name(changelog_format),
                MLModel.by_name(changelog_classifier),
            )
        else:
            changelog = repo.git.log(
                f"{old_version}..HEAD", no_merges=True, format="* %s"
            ).splitlines()

        if version_file:
            # TODO: We should prepend changes instead of appending them.
            _LOGGER.info("Adding changelog to the CHANGELOG.md file")
            with open("CHANGELOG.md", "a") as changelog_file:
                changelog_file.write(
                    f"\n## Release {new_version} ({datetime.now().replace(microsecond=0).isoformat()})\n"
                )
                changelog_file.write("\n".join(changelog))
                changelog_file.write("\n")
            repo.git.add("CHANGELOG.md")

        _LOGGER.info("Computed changelog has %d entries", len(changelog))
        return changelog

    @staticmethod
    def _adjust_pr_body(issue: Issue) -> str:
        if not issue.description:
            return ""

        result = "\n".join(issue.description.splitlines())
        result = result.replace(
            "Hey, Kebechet!\n\nCreate a new patch release, please.",
            f"Hey, @{issue.author}!\n\nOpening this PR to fix the last release.\n\n",
        )

        result = result.replace(
            "Hey, Kebechet!\n\nCreate a new minor release, please.",
            f"Hey, @{issue.author}!\n\nOpening this PR to create a release in a backwards compatible manner.\n\n",
        )

        return result.replace(
            "Hey, Kebechet!\n\nCreate a new major release, please.",
            f"Hey, @{issue.author}!\n\nYour possible backwards incompatible changes will be released by this PR.\n\n",
        )

    @classmethod
    def _construct_pr_body(cls, issue: Issue, changelog: typing.List[str]) -> str:
        """Construct body of the opened pull request with version update."""
        # Copy body from the original issue, this is helpful in case of
        # instrumenting CI (e.g. Depends-On in case of Zuul) so automatic
        # merges are perfomed as desired.
        body = cls._adjust_pr_body(issue)
        truncated_changelog = changelog[:_MAX_CHANELOG_SIZE]
        body += (
            "Related: #"
            + str(issue.id)
            + "\n\n```"
            + "\n\nChangelog:\n"
            + "\n".join(truncated_changelog)
            + "\n```"
        )
        if len(changelog) > _MAX_CHANELOG_SIZE:
            body += "\n" + _BODY_TRUNCATED
        return body

    def run(  # type: ignore
        self,
        maintainers: list = None,
        assignees: list = None,
        labels: list = None,
        changelog_file: bool = False,
        changelog_smart: bool = True,
        changelog_classifier: str = MLModel.DEFAULT.name,
        changelog_format: str = Format.DEFAULT.name,
    ) -> None:
        """Check issues for new issue request, if a request exists, issue a new PR with adjusted version in sources."""
        if self.parsed_payload:
            if self.parsed_payload.get("event") not in _EVENTS_SUPPORTED:
                _LOGGER.info(
                    "Version Manager doesn't act on %r events.",
                    self.parsed_payload.get("event"),
                )
                return

        reported_issues = []
        for issue in self.sm.repository.get_issue_list():
            issue_title = issue.title.strip()

            if issue_title.startswith(
                (_NO_VERSION_FOUND_ISSUE_NAME, _MULTIPLE_VERSIONS_FOUND_ISSUE_NAME)
            ):
                # Reported issues that should be closed on success version change.
                reported_issues.append(issue)

            # This is an optimization not to clone repo each time.
            if not self._is_release_request(issue_title):
                continue

            _LOGGER.info(
                "Found an issue #%s which is a candidate for request of new version release: %s",
                issue.id,
                issue.title,
            )

            with cloned_repo(self.service_url, self.slug) as repo:
                if assignees:
                    try:
                        self.sm.assign(issue, assignees)
                    except Exception:
                        _LOGGER.exception(
                            f"Failed to assign {assignees} to issue #{issue.id}"
                        )
                        issue.comment(
                            "Unable to assign provided assignees, please check bot configuration."
                        )

                maintainers = maintainers or self._get_maintainers(labels)
                if issue.author.lower() not in (m.lower() for m in maintainers):
                    issue.comment(
                        f"Sorry, @{issue.author} but you are not stated in maintainers section for "
                        f"this project. Maintainers are @" + ", @".join(maintainers)
                        if maintainers
                        else "Sorry, no maintainers configured."
                    )
                    issue.close()
                    # Next issue.
                    continue

                try:
                    version_identifier, old_version = self._adjust_version_in_sources(  # type: ignore
                        repo, labels, issue
                    )
                except VersionError as exc:
                    _LOGGER.exception("Failed to adjust version information in sources")
                    issue.comment(str(exc))
                    issue.close()
                    raise

                if not version_identifier:
                    _LOGGER.error("Giving up with automated release")
                    return

                try:
                    changelog = self._compute_changelog(
                        repo,
                        old_version,
                        version_identifier,
                        changelog_smart,
                        changelog_classifier,
                        changelog_format,
                        version_file=changelog_file,
                    )
                except ThothGlyphException as exc:
                    _LOGGER.exception("Failed to generate smart release log")
                    issue.comment(str(exc))
                    issue.close()
                    return

                # If an issue exists, we close it as there is no change to source code.
                if not changelog:
                    message = f"Closing the issue as there is no changelog between the new release of {self.slug}."
                    _LOGGER.info(message)
                    issue.comment(message)
                    issue.close()
                    return

                branch_name = "v" + version_identifier
                repo.git.checkout("HEAD", b=branch_name)
                message = _VERSION_PULL_REQUEST_NAME.format(version_identifier)
                repo.index.commit(message)
                # If this PR already exists, this will fail.
                repo.remote().push(branch_name)

                request = self.sm.open_merge_request(
                    message,
                    branch_name,
                    body=self._construct_pr_body(issue, changelog),
                    labels=labels,
                )

                _LOGGER.info(
                    f"Opened merge request with {request.id} for new release of {self.slug} "
                    f"in version {version_identifier}"
                )

        for reported_issue in reported_issues:
            reported_issue.comment("Closing as this issue is no longer relevant.")
            reported_issue.close()



#!/usr/bin/env python3
# Kebechet
# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Dependency update management logic."""

import os
import logging
import toml
import re
import json
import typing
from itertools import chain
from functools import partial

import git

from kebechet.exception import DependencyManagementError
from kebechet.exception import InternalError
from kebechet.exception import PipenvError
from kebechet.managers.manager import ManagerBase
from thoth.sourcemanagement import Issue
from thoth.sourcemanagement import PullRequest
from thoth.sourcemanagement import PRStatus
from kebechet.utils import cloned_repo

from .messages import ISSUE_CLOSE_COMMENT
from .messages import ISSUE_COMMENT_UPDATE_ALL
from .messages import ISSUE_INITIAL_LOCK
from .messages import ISSUE_NO_DEPENDENCY_MANAGEMENT
from .messages import ISSUE_PIPENV_UPDATE_ALL
from .messages import ISSUE_REPLICATE_ENV
from .messages import ISSUE_UNSUPPORTED_PACKAGE
from kebechet.utils import construct_raw_file_url

_LOGGER = logging.getLogger(__name__)
_RE_VERSION_DELIMITER = re.compile("(==|===|<=|>=|~=|!=|<|>|\\[)")

_ISSUE_UPDATE_ALL_NAME = "Failed to update dependencies to their latest version"
_ISSUE_INITIAL_LOCK_NAME = "Failed to perform initial lock of software stack"
_ISSUE_REPLICATE_ENV_NAME = "Failed to replicate environment for updates"
_ISSUE_NO_DEPENDENCY_NAME = "No dependency management found"
_ISSUE_UNSUPPORTED_PACKAGE = (
    "Application cannot be managed by Kebechet due to Git package"
)
_ISSUE_MANUAL_UPDATE = "Kebechet update"

# Github and Gitlab events on which the manager acts upon.
_EVENTS_SUPPORTED = ["push", "issues", "issue", "merge_request"]

# Note: We cannot use pipenv as a library (at least not now - version 2018.05.18) - there is a need to call it
# as a subprocess as pipenv keeps path to the virtual environment in the global context that is not
# updated on subsequent calls.


class UpdateManager(ManagerBase):
    """Manage updates of dependencies."""

    def __init__(self, *args, **kwargs):
        """Initialize update manager."""
        self._repo = None
        # We do API calls once for merge requests and we cache them for later use.
        self._cached_merge_requests = None
        super().__init__(*args, **kwargs)

    @property
    def repo(self):
        """Get repository on which we work on."""
        return self._repo

    @repo.setter
    def repo(self, repo: git.Repo):
        """Set repository information and all derived information needed."""
        self._repo = repo
        self.slug = repo.remote().url.split(":", maxsplit=1)[1][: -len(".git")]

    @property
    def sha(self):
        """Get SHA of the current head commit."""
        return self.repo.head.commit.hexsha

    @staticmethod
    def _get_dependency_version(dependency: str, is_dev: bool) -> str:
        """Get version of the given dependency from Pipfile.lock."""
        try:
            with open("Pipfile.lock") as pipfile_lock:
                pipfile_lock_content = json.load(pipfile_lock)
        except Exception as exc:
            # TODO: open a PR to fix this
            raise DependencyManagementError(
                f"Failed to load Pipfile.lock file: {str(exc)}"
            ) from exc

        # We look for normalized dependency in Pipfile.lock.
        normalized_dependency = re.sub(r"[-_.]+", "-", dependency).lower()
        version = (
            pipfile_lock_content["develop" if is_dev else "default"]
            .get(normalized_dependency, {})
            .get("version")
        )
        # if not kept as normaized depedency in Pipfile.lock.
        version = (
            pipfile_lock_content["develop" if is_dev else "default"]
            .get(dependency, {})
            .get("version")
        )
        if not version:
            raise InternalError(
                f"Failed to retrieve version information for dependency {dependency}, (dev: {is_dev})"
            )

        return version[len("==") :]

    @staticmethod
    def _get_direct_dependencies() -> tuple:
        """Get all direct dependencies stated in the Pipfile file."""
        try:
            pipfile_content = toml.load("Pipfile")
        except Exception as exc:
            # TODO: open a PR to fix this
            raise DependencyManagementError(
                f"Failed to load Pipfile: {str(exc)}"
            ) from exc

        default = list(
            package_name.lower() for package_name in pipfile_content["packages"].keys()
        )
        develop = list(
            package_name.lower()
            for package_name in pipfile_content.get("dev-packages", {}).keys()
        )

        return default, develop

    @staticmethod
    def _get_direct_dependencies_requirements(req_dev: bool) -> set:
        """Gather all direct dependencies.

        Get all direct dependencies based on either requirements.in or requirements-dev.in file
        and generated Pipfile.lock from it.
        """
        input_file = "requirements-dev.in" if req_dev else "requirements.in"
        with open(input_file, "r") as requirements_in_file:
            content = requirements_in_file.read()

        direct_dependencies = set()
        for line in content.splitlines():
            if line.strip().startswith("#"):
                continue

            # TODO: we could reuse pip or pipenv functionality here to parse file.
            package_name = _RE_VERSION_DELIMITER.split(line)[0]
            direct_dependencies.add(package_name.lower())

        return direct_dependencies

    def _get_all_packages_versions(self) -> dict:
        """Parse Pipfile.lock file and retrieve all packages in the corresponding locked versions."""
        try:
            with open("Pipfile.lock") as pipfile_lock:
                pipfile_lock_content = json.load(pipfile_lock)
        except Exception as exc:
            # TODO: open a PR to fix this
            raise DependencyManagementError(
                f"Failed to load Pipfile.lock file: {str(exc)}"
            ) from exc

        result = {}

        for package_name, package_info in pipfile_lock_content["default"].items():
            if "git" in package_info:
                self._create_unsupported_package_issue(package_name)
                raise DependencyManagementError(
                    f"Failed to find version in package that uses git source."
                )
            result[package_name.lower()] = {
                "dev": False,
                "version": package_info["version"][len("==") :],
            }

        for package_name, package_info in pipfile_lock_content["develop"].items():
            if "git" in package_info:
                self._create_unsupported_package_issue(package_name)
                raise DependencyManagementError(
                    "Failed to find version in package that uses git source."
                )
            result[package_name.lower()] = {
                "dev": False,
                "version": package_info["version"][len("==") :],
            }
        # Close git as a source issues.
        self.sm.close_issue_if_exists(
            _ISSUE_UNSUPPORTED_PACKAGE,
            f"Issue closed as no packages use git as a \
                    source anymore. Related SHA - {self.sha}",
        )
        return result

    def _create_unsupported_package_issue(self, package_name):
        """Create an issue as Kebechet doesn't support packages with git as source."""
        _LOGGER.info("Key Errror encountered, due package source being git.")
        pip_url = construct_raw_file_url(
            self.service_url, self.slug, "Pipfile", self.service_type
        )
        piplock_url = construct_raw_file_url(
            self.service_url, self.slug, "Pipfile.lock", self.service_type
        )
        self.sm.open_issue_if_not_exist(
            _ISSUE_UNSUPPORTED_PACKAGE,
            lambda: ISSUE_UNSUPPORTED_PACKAGE.format(
                sha=self.sha,
                package=package_name,
                pip_url=pip_url,
                piplock_url=piplock_url,
                environment_details=self.get_environment_details(),
            ),
        )

    @classmethod
    def _get_direct_dependencies_version(cls) -> dict:
        """Get versions of all direct dependencies based on the currently present Pipfile.lock."""
        default, develop = cls._get_direct_dependencies()

        result = {}
        default, develop = (
            ((dep, False) for dep in default),
            ((dep, True) for dep in develop),
        )
        for dependency, is_dev in chain(default, develop):
            version = cls._get_dependency_version(dependency, is_dev=is_dev)
            result[dependency] = {"version": version, "dev": is_dev}

        return result

    @staticmethod
    def _get_requirements_txt_dependencies(req_dev: bool) -> dict:
        """Gather dependencies from fully pinned down stack.

        Gather dependencies from either requirements.txt or requirements-dev.txt file,
        our requirements.txt and requirements-dev.txt holds fully pinned down stack.
        """
        result = {}
        input_file = "requirements-dev.txt" if req_dev else "requirements.txt"
        with open(input_file, "r") as requirements_file:
            content = requirements_file.read()

        for line in content.splitlines():
            line = line.strip()

            if line.startswith(("#", "-")) or not line:
                continue

            package_and_version = line.split("==", maxsplit=1)
            if len(package_and_version) != 2:
                raise DependencyManagementError(
                    f"File {input_file} does not state fully locked "
                    f"dependencies: {line!r} is not fully qualified dependency"
                )
            package_name, package_version = package_and_version
            result[package_name] = {
                # FIXME: tabs?
                "version": package_version.split(r" ", maxsplit=1)[0],
                "dev": False,
            }

        return result

    @staticmethod
    def _construct_branch_name(package_name: str, new_package_version: str) -> str:
        """Construct branch name for the updated dependency."""
        return f"kebechet-{package_name}-{new_package_version}"

    def _open_merge_request_update(
        self,
        dependency: str,
        old_version: str,
        new_version: str,
        labels: typing.Optional[list],
        files: list,
        merge_request: PullRequest,
    ) -> typing.Optional[int]:
        """Open a pull/merge request for dependency update."""
        branch_name = self._construct_branch_name(dependency, new_version)
        commit_msg = f"Automatic update of dependency {dependency} from {old_version} to {new_version}"

        # If we have already an update for this package we simple issue git
        # push force always to keep branch up2date with the recent master and avoid merge conflicts.
        self._git_push(":pushpin: " + commit_msg, branch_name, files, force_push=True)

        if not merge_request:
            _LOGGER.info(
                f"Creating a pull request to update {dependency} from version {old_version} to {new_version}"
            )
            body = (
                f"Dependency {dependency} was used in version {old_version}, "
                f"but the current latest version is {new_version}."
            )
            merge_request = self.sm.open_merge_request(
                commit_msg, branch_name, body, labels
            )
            return merge_request

        _LOGGER.info(
            f"Pull request #{merge_request.id} to update {dependency} from "
            f"version {old_version} to {new_version} updated"
        )
        merge_request.comment(
            f"Pull request has been rebased on top of the current master with SHA {self.sha}"
        )
        return merge_request

    def _should_update(self, package_name, new_package_version) -> tuple:
        """Check whether the given update was already proposed as a pull request."""
        branch_name = self._construct_branch_name(package_name, new_package_version)
        response = {
            mr
            for mr in self._cached_merge_requests
            if mr.source_branch == branch_name and mr.status == PRStatus.open
        }

        if len(response) == 0:
            _LOGGER.debug(
                f"No pull request was found for update of {package_name} to version {new_package_version}"
            )
            return None, True
        elif len(response) == 1:
            response = list(response)[0]
            commits = response.get_all_commits()  # type: ignore
            if len(commits) != 1:
                _LOGGER.info(
                    "Update of package {package_name} to version {new_package_version} will not be issued,"
                    "the pull request as additional commits (by a maintaner?)"
                )
                return response, False

            pr_number = response.id  # type: ignore
            if self.sha != commits[0]:
                _LOGGER.debug(
                    f"Found already existing  pull request #{pr_number} for old master "
                    f"branch {commits[0][:7]!r} updating pull request based on "
                    f"branch {branch_name!r} for the current master branch {self.sha[:7]!r}"
                )
                return response, True
            else:
                _LOGGER.debug(
                    f"Found already existing  pull request #{pr_number} for the current master "
                    f"branch {self.sha[:7]!r}, not updating pull request"
                )
                return response, False
        else:
            raise InternalError(
                f"Multiple ({len(response)}) pull requests with same "
                f"branch name {branch_name!r} opened."
            )

    def _git_push(
        self, commit_msg: str, branch_name: str, files: list, force_push: bool = False
    ) -> None:
        """Perform git push after adding files and giving a commit message."""
        self.repo.git.checkout("HEAD", b=branch_name)
        self.repo.index.add(files)
        self.repo.index.commit(commit_msg)
        self.repo.remote().push(branch_name, force=force_push)

    def _get_all_outdated(self, old_direct_dependencies: dict) -> dict:
        """Get all outdated packages based on Pipfile.lock."""
        new_direct_dependencies = self._get_direct_dependencies_version()

        result = {}
        for package_name in old_direct_dependencies.keys():
            if old_direct_dependencies[package_name][
                "version"
            ] != new_direct_dependencies.get(package_name, {}).get("version"):
                old_version = old_direct_dependencies[package_name]["version"]
                new_version = new_direct_dependencies.get(package_name, {}).get(
                    "version"
                )
                is_dev = old_direct_dependencies[package_name]["dev"]

                _LOGGER.debug(
                    f"Found new update for {package_name}: {old_version} -> {new_version} (dev: {is_dev})"
                )
                result[package_name] = {
                    "dev": is_dev,  # This should not change
                    "old_version": old_version,
                    "new_version": new_version,
                }

        return result

    @classmethod
    def _pipenv_lock_requirements(cls, output_file: str) -> None:
        """Perform pipenv lock into requirements.txt or requirements-dev.txt file."""
        result = cls.run_pipenv("pipenv lock -r ")
        with open(output_file, "w") as requirements_file:
            # Remove markers from requirements file.
            for line in result.splitlines():
                parts = line.split(" ; ", maxsplit=1)
                requirements_file.write(parts[0])

    def _create_update(
        self,
        dependency: str,
        package_version: str,
        old_version: str,
        is_dev: bool = False,
        labels: list = None,
        old_environment: dict = None,
        merge_request: PullRequest = None,
        pipenv_used: bool = True,
        req_dev: bool = False,
    ) -> typing.Union[tuple, None]:
        """Create an update for the given dependency when dependencies are managed by Pipenv.

        The old environment is set to a non None value only if we are operating on requirements.{in,txt}. It keeps
        information of packages that were present in the old environment so we can selectively change versions in the
        already existing requirements.txt or add packages that were introduced as a transitive dependency.
        """
        cmd = f"pipenv install {dependency}=={package_version} --keep-outdated"
        if is_dev:
            cmd += " --dev"
        self.run_pipenv(cmd)
        if pipenv_used:
            # Discard changes by pipenv made in Pipfile (dependency lock) as it affects hashes computed for
            # Pipfile.lock. We don't do `pipenv update` as in some cases pipenv does not update dependencies at all.
            self.repo.git.checkout("--", "Pipfile")
        self.run_pipenv("pipenv lock --keep-outdated")

        if not old_environment:
            merge_request = self._open_merge_request_update(
                dependency,
                old_version,
                package_version,
                labels,
                ["Pipfile.lock"],
                merge_request,
            )
            return old_version, package_version, merge_request.id  # type: ignore

        # For either requirements.txt  or requirements-dev.text scenario we need to propagate all changes
        # (updates of transitive dependencies) into requirements.txt or requirements-dev file
        output_file = "requirements-dev.txt" if req_dev else "requirements.txt"
        self._pipenv_lock_requirements(output_file)
        merge_request = self._open_merge_request_update(
            dependency,
            old_version,
            package_version,
            labels,
            [output_file],
            merge_request,
        )
        return old_version, package_version, merge_request.id  # type: ignore

    @classmethod
    def _replicate_old_environment(cls) -> None:
        """Replicate old environment based on its specification - packages in specific versions."""
        _LOGGER.info("Replicating old environment for incremental update")
        cls.run_pipenv("pipenv sync --dev")

    @classmethod
    def _create_pipenv_environment(cls, input_file: str) -> None:
        """Create a pipenv environment - Pipfile and Pipfile.lock from requirements.in or requirements-dev.in file."""
        if os.path.isfile(input_file) and input_file == "requirements.in":
            _LOGGER.info(f"Installing dependencies from {input_file}")
            cls.run_pipenv(f"pipenv install -r {input_file}")
        elif os.path.isfile(input_file) and input_file == "requirements-dev.in":
            _LOGGER.info(f"Installing dependencies from {input_file}")
            cls.run_pipenv(f"pipenv install -r {input_file} --dev")
        else:
            raise DependencyManagementError(
                "No dependency management found in the repo - no Pipfile nor requirements.in nor requirements-dev.in"
            )

    def _get_prs(self, source_branch_name: str) -> set:
        """Get pull requests with source branch name."""
        return {
            mr
            for mr in self.sm.get_prs()
            if mr.source_branch == source_branch_name and mr.status == PRStatus.open
        }

    def _create_initial_lock(
        self, labels: list, pipenv_used: bool, req_dev: bool
    ) -> bool:
        """Perform initial requirements lock into requirements.txt file."""
        # We use lock_func to optimize run - it will be called only if actual locking needs to be performed.
        if not pipenv_used and not os.path.isfile("requirements.txt") and not req_dev:
            _LOGGER.info("Initial lock based on requirements.in will be done")
            lock_func = partial(self._pipenv_lock_requirements, "requirements.txt")
        elif not pipenv_used and not os.path.isfile("requirements-dev.txt") and req_dev:
            _LOGGER.info("Initial lock based on requirements-dev.in will be done")
            lock_func = partial(self._pipenv_lock_requirements, "requirements-dev.txt")
        elif pipenv_used and not os.path.isfile("Pipfile.lock"):
            _LOGGER.info("Initial lock based on Pipfile will be done")
            lock_func = partial(self.run_pipenv, "pipenv lock")
        else:
            return False

        branch_name = "kebechet-initial-lock"
        request = self._get_prs(source_branch_name=branch_name)

        if req_dev and not pipenv_used:
            files = ["requirements-dev.txt"]
        elif not req_dev and not pipenv_used:
            files = ["requirements.txt"]
        else:
            files = ["Pipfile.lock"]

        commit_msg = "Initial dependency lock"
        if len(request) == 0:
            lock_func()
            self._git_push(commit_msg, branch_name, files)
            request = self.sm.open_merge_request(
                commit_msg, branch_name, body="", labels=labels
            )
            _LOGGER.info(
                f"Initial dependency lock present in PR #{request.id}"  # type: ignore
            )
        elif len(request) == 1:
            request = list(request)[0]
            commits = request.get_all_commits()  # type: ignore

            if len(commits) != 1:
                _LOGGER.info(
                    "There have been done changes in the original pull request (multiple commits found), "
                    "aborting doing changes to the adjusted opened pull request"
                )
                return False

            if self.sha != commits[0]:
                lock_func()
                self._git_push(commit_msg, branch_name, files, force_push=True)
                request.comment(  # type: ignore
                    f"Pull request has been rebased on top of the current master with SHA {self.sha}"
                )
            else:
                _LOGGER.info(
                    f"Pull request #{request.id} is up to date for the current master branch"  # type: ignore
                )
        else:
            raise DependencyManagementError(
                f"Found two or more pull requests for initial requirements lock for branch {branch_name}"
            )

        return True

    @classmethod
    def _pipenv_update_all(cls):
        """Update all dependencies to their latest version."""
        _LOGGER.info("Updating all dependencies to their latest version")
        cls.run_pipenv("pipenv update --dev")
        cls.run_pipenv("pipenv lock")
        return None

    def _add_refresh_comment(
        self, exc: PipenvError, issue: Issue
    ) -> typing.Optional[str]:
        """Create a refresh comment to an issue if the given master has some changes."""
        if self.sha in issue.description:
            _LOGGER.debug("No need to update refresh comment, the issue is up to date")
            return None

        for issue_comment in issue.get_comments():
            if self.sha in issue_comment.body:
                _LOGGER.debug(
                    f"No need to update refresh comment, comment for the current "
                    f"master {self.sha[:7]!r} found in a comment"
                )
                break
        else:
            return ISSUE_COMMENT_UPDATE_ALL.format(
                sha=self.sha,
                slug=self.slug,
                environment_details=self.get_environment_details(),
                dependency_graph=self.get_dependency_graph(graceful=True),
                **exc.__dict__,
            )
        return None

    def _relock_all(self, exc: PipenvError, labels: list) -> None:
        """Re-lock all dependencies given the Pipfile."""
        pip_url = construct_raw_file_url(
            self.service_url, self.slug, "Pipfile", self.service_type
        )
        piplock_url = construct_raw_file_url(
            self.service_url, self.slug, "Pipfile.lock", self.service_type
        )
        issue = self.sm.open_issue_if_not_exist(
            _ISSUE_REPLICATE_ENV_NAME,
            lambda: ISSUE_REPLICATE_ENV.format(
                **exc.__dict__,
                sha=self.sha,
                pip_url=pip_url,
                piplock_url=piplock_url,
                environment_details=self.get_environment_details(),
            ),
            refresh_comment=partial(self._add_refresh_comment, exc),
            labels=labels,
        )

        self._pipenv_update_all()
        commit_msg = "Automatic dependency re-locking"
        branch_name = "kebechet-dependency-relock"

        existing_prs = self._get_prs(branch_name)
        if len(existing_prs) == 1:
            pr = list(existing_prs)[0]
            commits = pr.get_all_commits()
            if len(commits) != 1:
                pr.comment(
                    "There have been done changes in the original pull request (multiple commits found), "
                    "aborting doing changes to the modified opened pull request"
                )
                return None
            if self.sha != commits[0]:
                self._git_push(
                    ":pushpin: " + commit_msg,
                    branch_name,
                    ["Pipfile.lock"],
                    force_push=True,
                )
                pr.comment(
                    f"Pull request has been rebased on top of the current master with SHA {self.sha}"
                )
        elif len(existing_prs) == 0:
            # Default case
            self._git_push(":pushpin: " + commit_msg, branch_name, ["Pipfile.lock"])
            pr_id = self.sm.open_merge_request(
                commit_msg, branch_name, f"Fixes: #{issue.id}", labels
            )
            _LOGGER.info(
                f"Issued automatic dependency re-locking in PR #{pr_id} to fix issue #{issue.id}"
            )
        else:
            raise DependencyManagementError(
                f"Found two or more pull requests for automatic relock for branch {branch_name}"
            )

    def _delete_old_branches(self, outdated: dict) -> None:
        """Delete old kebechet branches from the remote repository."""
        branches = {
            entry for entry in self.sm.list_branches() if entry.startswith("kebechet-")
        }
        for package_name, info in outdated.items():
            # Do not remove active branches - branches we issued PRs in.
            branch_name = self._construct_branch_name(package_name, info["new_version"])
            try:
                branches.remove(branch_name)
            except KeyError:
                # e.g. if there was an issue with PR opening.
                pass

        for branch_name in branches:
            _LOGGER.debug(f"Deleting old branch {branch_name}")
            try:
                self.sm.delete_branch(branch_name)
            except Exception:
                _LOGGER.exception(f"Failed to delete inactive branch {branch_name}")

    def _do_update(
        self, labels: list, pipenv_used: bool = False, req_dev: bool = False
    ) -> dict:
        """Update dependencies based on management used."""
        close_initial_lock_issue = partial(
            self.sm.close_issue_if_exists,
            _ISSUE_INITIAL_LOCK_NAME,
            comment=ISSUE_CLOSE_COMMENT.format(sha=self.sha),
        )

        # Check for first time (initial) locks first.
        try:
            if self._create_initial_lock(labels, pipenv_used, req_dev):
                close_initial_lock_issue()
                return {}
        except PipenvError as exc:
            _LOGGER.exception("Failed to perform initial dependency lock")
            file_name = (
                "requirements.txt"
                if not req_dev
                else "requirements-dev.txt"
                if not pipenv_used
                else "Pipfile"
            )
            file_url = construct_raw_file_url(
                self.service_url, self.slug, file_name, self.service_type
            )
            self.sm.open_issue_if_not_exist(
                _ISSUE_INITIAL_LOCK_NAME,
                body=lambda: ISSUE_INITIAL_LOCK.format(
                    sha=self.sha,
                    url=file_url,
                    file=file_name,
                    environment_details=self.get_environment_details(),
                    **exc.__dict__,  # noqa F821
                ),
                refresh_comment=partial(self._add_refresh_comment, exc),
                labels=labels,
            )
            raise

        close_initial_lock_issue()

        if pipenv_used:
            old_environment = self._get_all_packages_versions()
            old_direct_dependencies_version = self._get_direct_dependencies_version()
            try:
                self._pipenv_update_all()
            except PipenvError as exc:
                _LOGGER.warning(
                    "Failed to update dependencies to their latest version, reporting issue"
                )
                pip_url = construct_raw_file_url(
                    self.service_url, self.slug, "Pipfile", self.service_type
                )
                piplock_url = construct_raw_file_url(
                    self.service_url, self.slug, "Pipfile.lock", self.service_type
                )
                self.sm.open_issue_if_not_exist(
                    _ISSUE_UPDATE_ALL_NAME,
                    body=lambda: ISSUE_PIPENV_UPDATE_ALL.format(
                        sha=self.sha,
                        pip_url=pip_url,
                        piplock_url=piplock_url,
                        environment_details=self.get_environment_details(),
                        dependency_graph=self.get_dependency_graph(graceful=True),
                        **exc.__dict__,  # noqa F821
                    ),
                    refresh_comment=partial(self._add_refresh_comment, exc),
                    labels=labels,
                )
                return {}
            else:
                # We were able to update all, close reported issue if any.
                self.sm.close_issue_if_exists(
                    _ISSUE_UPDATE_ALL_NAME,
                    comment=ISSUE_CLOSE_COMMENT.format(sha=self.sha),
                )
        else:  # either requirements.txt or requirements-dev.txt
            old_environment = self._get_requirements_txt_dependencies(req_dev)
            direct_dependencies = self._get_direct_dependencies_requirements(req_dev)
            old_direct_dependencies_version = {
                k: v for k, v in old_environment.items() if k in direct_dependencies
            }

        outdated = self._get_all_outdated(old_direct_dependencies_version)
        _LOGGER.info(f"Outdated: {outdated}")

        # Undo changes made to Pipfile.lock by _pipenv_update_all.
        self.repo.head.reset(index=True, working_tree=True)

        result = {}
        if outdated:
            # Do API calls only once, cache results.
            self._cached_merge_requests = self.sm.get_prs()

        for package_name in outdated.keys():
            # As an optimization, first check if the given PR is already present.
            new_version = outdated[package_name]["new_version"]
            old_version = outdated[package_name]["old_version"]

            merge_request, should_update = self._should_update(
                package_name, new_version
            )
            if not should_update:
                _LOGGER.info(
                    f"Skipping update creation for {package_name} from version {old_version} to "
                    f"{new_version} as the given update already exists in PR #{merge_request.id}"
                )
                continue

            try:
                self._replicate_old_environment()
            except PipenvError as exc:
                # There has been an error in locking dependencies. This can be due to a missing dependency or simply
                # currently locked dependencies are not correct. Try to issue a pull request that would fix
                # that. We know that update all works, use update.
                _LOGGER.warning(
                    "Failed to replicate old environment, re-locking all dependencies"
                )
                self._relock_all(exc, labels)
                return {}

            is_dev = outdated[package_name]["dev"]
            try:
                _LOGGER.info(
                    f"Creating update of dependency {package_name} in repo {self.slug} (devel: {is_dev})"
                )
                versions = self._create_update(
                    package_name,
                    new_version,
                    old_version,
                    is_dev=is_dev,
                    labels=labels,
                    old_environment=old_environment if not pipenv_used else None,
                    merge_request=merge_request,
                    pipenv_used=pipenv_used,
                    req_dev=req_dev,
                )
                if versions:
                    result[package_name] = versions
            except Exception as exc:
                _LOGGER.exception(
                    f"Failed to create update for dependency {package_name}: {str(exc)}"
                )
            finally:
                self.repo.head.reset(index=True, working_tree=True)
                self.repo.git.checkout("master")

        # We know that locking was done correctly - if the issue is still open, close it. The issue
        # should be automatically closed by merging the generated PR.
        self.sm.close_issue_if_exists(
            _ISSUE_REPLICATE_ENV_NAME, comment=ISSUE_CLOSE_COMMENT.format(sha=self.sha)
        )

        self._delete_old_branches(outdated)
        return result

    def run(self, labels: list) -> typing.Optional[dict]:
        """Create a pull request for each and every direct dependency in the given org/repo (slug)."""
        if self.parsed_payload:
            if self.parsed_payload.get("event") not in _EVENTS_SUPPORTED:
                _LOGGER.info(
                    "Update Manager doesn't act on %r events.",
                    self.parsed_payload.get("event"),
                )
                return None

        # We will keep venv in the project itself - we have permissions in the cloned repo.
        os.environ["PIPENV_VENV_IN_PROJECT"] = "1"

        with cloned_repo(self.service_url, self.slug, depth=1) as repo:
            # Make repo available in the instance.
            self.repo = repo

            close_no_management_issue = partial(
                self.sm.close_issue_if_exists,
                _ISSUE_NO_DEPENDENCY_NAME,
                comment=ISSUE_CLOSE_COMMENT.format(sha=self.sha),
            )

            close_manual_update_issue = partial(
                self.sm.close_issue_if_exists,
                _ISSUE_MANUAL_UPDATE,
                comment=ISSUE_CLOSE_COMMENT.format(sha=self.sha),
            )

            if os.path.isfile("Pipfile"):
                _LOGGER.info("Using Pipfile for dependency management")
                close_no_management_issue()
                result = self._do_update(labels, pipenv_used=True, req_dev=False)
                close_manual_update_issue()
            elif os.path.isfile("requirements.in"):
                self._create_pipenv_environment(input_file="requirements.in")
                _LOGGER.info("Using requirements.in for dependency management")
                close_no_management_issue()
                result = self._do_update(labels, pipenv_used=False, req_dev=False)
                if os.path.isfile("requirements-dev.in"):
                    self._create_pipenv_environment(input_file="requirements-dev.in")
                    _LOGGER.info("Using requirements-dev.in for dependency management")
                    close_no_management_issue()
                    result = self._do_update(labels, pipenv_used=False, req_dev=True)
                close_manual_update_issue()
            else:
                _LOGGER.warning("No dependency management found")
                self.sm.open_issue_if_not_exist(
                    _ISSUE_NO_DEPENDENCY_NAME,
                    lambda: ISSUE_NO_DEPENDENCY_MANAGEMENT,
                    labels=labels,
                )
                return {}

            return result
